///////////////////////////////////////////////////////////////////////////////
// File: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
//
// Copyright 2014 JetByte Limited.
//
// This software is provided "as is" without a warranty of any kind. All
// express or implied conditions, representations and warranties, including
// any implied warranty of merchantability, fitness for a particular purpose
// or non-infringement, are hereby excluded. JetByte Limited and its licensors
// shall not be liable for any damages suffered by licensee as a result of
// using the software. In no event will JetByte Limited be liable for any
// lost revenue, profit or data, or for direct, indirect, special,
// consequential, incidental or punitive damages, however caused and regardless
// of the theory of liability, arising out of the use of or inability to use
// software, even if JetByte Limited has been advised of the possibility of
// such damages.
//
///////////////////////////////////////////////////////////////////////////////

#include "JetByteTools/Admin/Admin.h"

#include "SocketServer.h"

#include "JetByteTools/CoreTools/FileUtils.h"
#include "JetByteTools/CoreTools/StringUtils.h"
#include "JetByteTools/CoreTools/DebugTrace.h"
#include "JetByteTools/CoreTools/Exception.h"
#include "JetByteTools/CoreTools/StringConverter.h"
#include "JetByteTools/CoreTools/IIndexedOpaqueUserData.h"

#include "JetByteTools/SocketTools/IStreamSocket.h"
#include "JetByteTools/SocketTools/IAllocateSequencedStreamSockets.h"

#include "JetByteTools/StreamingTools/RTSPConnection.h"

#pragma hdrstop

///////////////////////////////////////////////////////////////////////////////
// Using directives
///////////////////////////////////////////////////////////////////////////////

using JetByteTools::Core::OutputEx;
using JetByteTools::Core::DumpData;
using JetByteTools::Core::CStringConverter;
using JetByteTools::Core::CException;
using JetByteTools::Core::IIndexedOpaqueUserData;
using JetByteTools::Core::ToString;

using JetByteTools::IO::IAllocateBuffers;
using JetByteTools::IO::IBuffer;
using JetByteTools::IO::CSmartBuffer;
using JetByteTools::IO::IIOPool;
using JetByteTools::IO::IBufferChain;

using JetByteTools::Socket::CFilteringStreamSocketServerEx;
using JetByteTools::Socket::IStreamSocket;
using JetByteTools::Socket::IAllocateSequencedStreamSockets;
using JetByteTools::Socket::IFullAddress;
using JetByteTools::Socket::IAddress;
using JetByteTools::Socket::ListenBacklog;
using JetByteTools::Socket::NoZeroByteRead;
using JetByteTools::Socket::ConnectionClosureReason;
using JetByteTools::Socket::CFlowControlStreamSocketConnectionFilter;
using JetByteTools::Socket::IMonitorStreamSocketFlowControl;
using JetByteTools::Socket::SocketBufferSize;

using JetByteTools::Streaming::IManageRTSPResources;
using JetByteTools::Streaming::CRTSPConnection;

using std::string;

///////////////////////////////////////////////////////////////////////////////
// Debugging
///////////////////////////////////////////////////////////////////////////////

//#define RTSP_SERVER_DEBUG
//#define RTSP_SERVER_FLOW_CONTROL_DEBUG

#ifdef RTSP_SERVER_DEBUG
#define DEBUG_OUTPUT(e) JetByteTools::Core::OutputEx(e);
#else
#define DEBUG_OUTPUT(e)
#endif

#ifdef RTSP_SERVER_FLOW_CONTROL_DEBUG
#define DEBUG_FLOW_CONTROL(e) JetByteTools::Core::OutputEx(e);
#else
#define DEBUG_FLOW_CONTROL(e)
#endif

///////////////////////////////////////////////////////////////////////////////
// CSocketServer
///////////////////////////////////////////////////////////////////////////////

CSocketServer::CSocketServer(
   const IManageRTSPResources::SupportedTransports supportedTransports,
   IManageRTSPResources &resourceManager,
   const bool terminateActiveSessionsOnConnectionClose,
   const CFlowControlStreamSocketConnectionFilter::BufferCount maxBuffers,
   const CFlowControlStreamSocketConnectionFilter::PendingWriteCount maxPendingWrites,
   const CFlowControlStreamSocketConnectionFilter::Percentage notificationPercentageStepSize,
   const SocketBufferSize recvBufferSize,
   const SocketBufferSize sendBufferSize,
   const IFullAddress &address,
   ListenBacklog listenBacklog,
   IIOPool &ioPool,
   IAllocateSequencedStreamSockets &socketAllocator,
   IAllocateBuffers &bufferAllocator)
   :  CFilteringStreamSocketServerEx(address, listenBacklog, *this, ioPool, socketAllocator, bufferAllocator, recvBufferSize, sendBufferSize, NoZeroByteRead),
      m_flowControlFilter(*this, *this, socketAllocator, bufferAllocator, maxBuffers, *this, maxPendingWrites, false, notificationPercentageStepSize),
      m_supportedTransports(supportedTransports),
      m_resourceManager(resourceManager),
      m_terminateActiveSessionsOnConnectionClose(terminateActiveSessionsOnConnectionClose),
      m_connectionIndex(socketAllocator.RequestUserDataSlot(_T("CSocketServer::m_connectionIndex")))
{

}

CSocketServer::~CSocketServer()
{
   try
   {
      CFilteringStreamSocketServerEx::WaitForShutdownToComplete();
   }
   JETBYTE_CATCH_AND_LOG_ALL_IN_DESTRUCTORS_IF_ENABLED
}

void CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket,
   const IAddress & /*address*/)
{
   DEBUG_OUTPUT(_T("OnConnectionEstablished"));

   const auto pConnection = new CRTSPConnection(socket, m_resourceManager, m_supportedTransports, m_terminateActiveSessionsOnConnectionClose);

   socket.SetUserPointer(m_connectionIndex, pConnection);

   socket.TryRead();
}

void CSocketServer::OnSocketReleased(
   IIndexedOpaqueUserData &userData)
{
   DEBUG_OUTPUT(_T("OnSocketReleased"));

   const auto pConnection = reinterpret_cast<CRTSPConnection *>(userData.GetUserPointer(m_connectionIndex));

   delete pConnection;
}

void CSocketServer::OnReadCompleted(
   IStreamSocket &socket,
   CSmartBuffer &buffer)
{
   try
   {
      DEBUG_OUTPUT(_T("ReadCompleted"));

      DEBUG_OUTPUT(_T("Incoming Data:\r\n") + CStringConverter::AtoT(reinterpret_cast<const char *>(buffer->GetMemory()), buffer->GetUsed()));

      DEBUG_OUTPUT(_T("Incoming Data:\r\n") + DumpData(buffer->GetMemory(), buffer->GetUsed(), 60, true));

      auto pConnection = reinterpret_cast<CRTSPConnection *>(socket.GetUserPointer(m_connectionIndex));

      pConnection->OnDataReceived(buffer);
   }
   catch(const CException &e)
   {
      OutputEx(_T("ReadCompleted - Exception - ") + e.GetDetails());
      socket.AbortConnection();
   }
   catch(...)
   {
      OutputEx(_T("ReadCompleted - Unexpected exception"));
      socket.AbortConnection();
   }
}

void CSocketServer::OnConnectionClientClose(
   IStreamSocket &socket)
{
   DEBUG_OUTPUT(_T("OnConnectionClientClose"));

   auto pConnection = reinterpret_cast<CRTSPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->AbortActiveSessions();
}

void CSocketServer::OnConnectionReset(
   IStreamSocket &socket,
   const DWORD lastError)
{
   (void)lastError;

   DEBUG_OUTPUT(_T("OnConnectionReset: ") + ToString(lastError));

   auto pConnection = reinterpret_cast<CRTSPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->AbortActiveSessions();
}

void CSocketServer::OnConnectionClosure(
   IStreamSocket &socket,
   const ConnectionClosureReason reason)
{
   (void)reason;

   DEBUG_OUTPUT(_T("OnConnectionClosure: ") + GetClosureReasonAsString(reason));

   auto pConnection = reinterpret_cast<CRTSPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->AbortActiveSessions();
}

bool CSocketServer::OnListFull(
   IStreamSocket &socket,
   const ConnectionId id,
   IMonitorStreamSocketFlowControl &monitor,
   const IBuffer &newBuffer,
   IBufferChain &buffers,
   IBufferChain &discardedBuffers)
{
   (void)id;
   (void)monitor;
   (void)newBuffer;
   (void)buffers;
   (void)discardedBuffers;

   DEBUG_FLOW_CONTROL(_T("OnListFull"));

   auto pConnection = reinterpret_cast<CRTSPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->PauseActiveSessions();

   // return true to add the buffer to the queue anyway and false to discard.
   // DO NOT ADD THE NEW BUFFER INTO THE DISCARDED BUFFERS LIST IF YOU DISCARD!!!

   return true;
}

void CSocketServer::OnListStatus(
   IStreamSocket &socket,
   const DWORD percentageUsed)
{
   (void)socket;
   (void)percentageUsed;

   DEBUG_FLOW_CONTROL(_T("OnListStatus - ") + ToString(percentageUsed) + _T("%"));
}

void CSocketServer::OnListEmpty(
   IStreamSocket &socket)
{
   DEBUG_FLOW_CONTROL(_T("OnListEmpty"));

   auto pConnection = reinterpret_cast<CRTSPConnection *>(socket.GetUserPointer(m_connectionIndex));

   pConnection->ResumeActiveSessions();
}

// Implement IMonitorStreamSocketFlowControl

IMonitorStreamSocketFlowControl::ConnectionId CSocketServer::OnConnectionEstablished(
   IStreamSocket &socket)
{
   const auto id = reinterpret_cast<IMonitorStreamSocketFlowControl::ConnectionId>(&socket);

   //DEBUG_FLOW_CONTROL(_T("OnConnectionEstablished: ") + ToString(id));

   return id;
}

void CSocketServer::OnWritePending(
   const ConnectionId id)
{
   (void)id;

   //DEBUG_FLOW_CONTROL(_T("OnWritePending: ") + ToString(id));
}

void CSocketServer::OnWriteComplete(
   const ConnectionId id)
{
   (void)id;

   //DEBUG_FLOW_CONTROL(_T("OnWriteComplete: ") + ToString(id));
}

bool CSocketServer::OnWriteCompletionError(
   const ConnectionId id,
   IStreamSocket &socket,
   IBuffer &buffer,
   const DWORD lastError)
{
   (void)id;
   (void)socket;
   (void)buffer;
   (void)lastError;

   //DEBUG_FLOW_CONTROL(_T("OnWriteCompletionError: ") + ToString(id) + _T(" - ") + ToString(lastError));

   // Return true if we can continue to send data and false if we should purge the flow control queue.

   return false;
}

void CSocketServer::OnBufferDelayed(
   const ConnectionId id,
   IBuffer &buffer)
{
   (void)buffer;
   (void)id;

   //DEBUG_FLOW_CONTROL(_T("OnBufferDelayed: ") + ToString(id));
}

void CSocketServer::OnConsolidation(
   const ConnectionId id,
   const bool consolidated)
{
   (void)consolidated;
   (void)id;
}

void CSocketServer::OnBufferConsolidated(
   const ConnectionId id,
   const bool wasDelayed)
{
   (void)wasDelayed;
   (void)id;
}

void CSocketServer::OnBufferSent(
   const ConnectionId id,
   IBuffer &buffer)
{
   (void)id;
   (void)buffer;

   //DEBUG_FLOW_CONTROL(_T("OnBufferSent: ") + ToString(id));
}

void CSocketServer::OnBufferDiscarded(
   const ConnectionId id,
   IBuffer &buffer,
   const bool wasDelayed)
{
   (void)id;
   (void)buffer;
   (void)wasDelayed;

   //DEBUG_FLOW_CONTROL(_T("OnBufferDiscarded: ") + ToString(id) + (wasDelayed ? _T(" was queued") : _T("")));
}

void CSocketServer::OnConnectionComplete(
   const ConnectionId id)
{
   (void)id;

   //DEBUG_FLOW_CONTROL(_T("OnConnectionComplete: ") + ToString(id));
}

void CSocketServer::OnConnectionTerminated(
   const ConnectionId id)
{
   (void)id;

   //DEBUG_FLOW_CONTROL(_T("OnConnectionTerminated: ") + ToString(id));
}

///////////////////////////////////////////////////////////////////////////////
// End of file: SocketServer.cpp
///////////////////////////////////////////////////////////////////////////////
